from unittest import TestCase
from semantic_text_similarity.data import load_sts_b_data
class TestData(TestCase):

    def test_load_train_dev_test_sts_b(self):
        train, dev, test = load_sts_b_data()
        return len(train) == 5749 and len(dev) == 1500 and len(test) == 1379


from pkg_resources import resource_exists, resource_listdir, resource_string, resource_stream,resource_filename

def load_sts_b_data():
    """
    Loads the STS-B dataset found here: http://ixa2.si.ehu.es/stswiki/index.php/STSbenchmark

    :return a tuple containing the train, dev, test split given by the data aggregators.
    """

    if not resource_exists('semantic_text_similarity', 'data/sts_b/sts-train.csv'):
        raise FileNotFoundError('Cannot find STS-B dataset')

    train = resource_string('semantic_text_similarity', 'data/sts_b/sts-train.csv').decode('utf-8').strip()
    dev = resource_string('semantic_text_similarity', 'data/sts_b/sts-dev.csv').decode('utf-8').strip()
    test = resource_string('semantic_text_similarity', 'data/sts_b/sts-dev.csv').decode('utf-8').strip()

    def yielder():
        for partition in (train, dev, test):
            data = []
            for idx,line in enumerate(partition.split('\n')):
                line = tuple(line.split("\t"))
                data.append({
                    'index': idx,
                    'sentence_1': line[5],
                    'sentence_2': line[6],
                    'similarity': float(line[4])
                })
            yield data

    return tuple([dataset for dataset in yielder()])


from pytorch_transformers import BertTokenizer
from pytorch_transformers.modeling_utils import WEIGHTS_NAME, CONFIG_NAME
from pytorch_transformers.modeling_bert import BertPreTrainedModel, BertConfig, BertModel

import torch
from torch import nn

from scipy.stats import pearsonr

from . import bert_sentence_pair_preprocessing

import logging, os

class BertSimilarityRegressor(BertPreTrainedModel):
    def __init__(self, bert_model_config: BertConfig):
        super(BertSimilarityRegressor, self).__init__(bert_model_config)
        self.bert = BertModel(bert_model_config)
        linear_size = bert_model_config.hidden_size
        if bert_model_config.pretrained_config_archive_map['additional_features'] is not None:
            linear_size+=bert_model_config.pretrained_config_archive_map['additional_features']

        self.regression = nn.Sequential(
            nn.Dropout(p=bert_model_config.hidden_dropout_prob),
            nn.Linear(linear_size, 1)
        )

        self.apply(self.init_weights)


    def forward(self, input_ids, token_type_ids, attention_masks, additional_features=None):
        """
        Feed forward network with one hidden layer.
        :param input_ids:
        :param token_type_ids:
        :param attention_masks:
        :return:
        """
        _,pooled_output = self.bert(input_ids, token_type_ids=token_type_ids, attention_mask=attention_masks)

        if additional_features is not None:
            pooled_output = torch.cat((pooled_output, additional_features),dim=1)

        return self.regression(pooled_output)


class BertSimilarity():
    """
    A class implementing the training and evaluation of a fine tuned BERT model for sentence pair similarity.
    """

    def __init__(self, args= None, device='cuda', bert_model_path='bert-base-uncased', batch_size=10, learning_rate = 5e-5, weight_decay=0, additional_features=None):
        if args is not None:
            self.args = vars(args)

        assert device in ['cuda', 'cpu']

        if not args:
            self.args = {}
            self.args['bert_model_path'] = bert_model_path
            self.args['device'] = device
            self.args['learning_rate'] = learning_rate
            self.args['weight_decay'] = weight_decay
            self.args['batch_size'] = batch_size

        self.log = logging.getLogger()



        self.bert_tokenizer = BertTokenizer.from_pretrained(self.args['bert_model_path'])
        if os.path.exists(self.args['bert_model_path']):
            if os.path.exists(os.path.join(self.args['bert_model_path'], CONFIG_NAME)):
                config = BertConfig.from_json_file(os.path.join(self.args['bert_model_path'], CONFIG_NAME))
            elif os.path.exists(os.path.join(self.args['bert_model_path'], 'bert_config.json')):
                config = BertConfig.from_json_file(os.path.join(self.args['bert_model_path'], 'bert_config.json'))
            else:
                raise ValueError("Cannot find a configuration for the BERT model you are attempting to load.")

        self.loss_function = torch.nn.MSELoss()

        config.pretrained_config_archive_map['additional_features'] = additional_features

        self.regressor_net = BertSimilarityRegressor.from_pretrained(self.args['bert_model_path'], config=config)
        self.optimizer = torch.optim.Adam(
            self.regressor_net.parameters(),
            weight_decay=self.args['weight_decay'],
            lr=self.args['learning_rate']
        )
        self.log.info('Initialized BertSentencePairSimilarity model from %s' % self.args['bert_model_path'])



    def predict(self, data: list, additional_features=None, return_predictions=True):
        """
        Given a list of sentence pair instances, makes predictions
        :param data:
        :param return predictions or log to model directory
        :return:
        """

        self.regressor_net.to(device=self.args['device'])
        self.regressor_net.eval()

        with torch.no_grad():

            if data is not None and isinstance(data, list):
                if isinstance(data[0], dict):
                    input_ids_eval, token_type_ids_eval, attention_masks_eval, correct_scores_eval = bert_sentence_pair_preprocessing(
                        data, self.bert_tokenizer)
                elif isinstance(data[0], tuple):
                    input_ids_eval, token_type_ids_eval, attention_masks_eval, correct_scores_eval = bert_sentence_pair_preprocessing(
                        [{'sentence_1': s1, 'sentence_2':s2} for s1, s2 in data], self.bert_tokenizer)
                else:
                    raise ValueError("Data must be a list of sentence pair tuples")

            predictions = torch.empty_like(correct_scores_eval)
            for i in range(0, input_ids_eval.shape[0], self.args['batch_size']):
                input_id_eval = input_ids_eval[i:i + self.args['batch_size']].to(device=self.args['device'])
                token_type_id_eval = token_type_ids_eval[i:i + self.args['batch_size']].to(device=self.args['device'])
                attention_mask_eval = attention_masks_eval[i:i + self.args['batch_size']].to(device=self.args['device'])

                if additional_features is not None:
                    additional_feature = additional_features[i:i + self.args['batch_size']].to(device=self.args['device'])
                else:
                    additional_feature = None

                predicted_score = self.regressor_net(input_id_eval, token_type_id_eval, attention_mask_eval, additional_features=additional_feature)
                predictions[i:i + self.args['batch_size']] = predicted_score


        if all(torch.isnan(correct_scores_eval)) or return_predictions: #no correct score labels are present, return the predictions
            return predictions.cpu().view(-1).numpy()
        else:
            self.log.info('Evaluating on Epoch %i' % (self.epoch))
            scores = {'pearson': 0}
            scores['pearson'] = \
                pearsonr(predictions.cpu().view(-1).numpy(), correct_scores_eval.view(-1).numpy())[0]

            with open(os.path.join(self.args['model_directory'], "eval_%s.csv" % self.epoch), 'w') as eval_results:
                eval_results.write("sentence_1\tsentence_2\tannotator_score\tpredicated_score\n")

                for idx, row in enumerate(torch.cat((correct_scores_eval, predictions.cpu()), dim=1)):
                    eval_results.write(
                        "%s\t%s\t%f\t%f\n" % (data[idx]['sentence_1'], data[idx]['sentence_2'], row[0].item(), row[1].item()))
                eval_results.write("Pearson\t%f" % (scores['pearson']))

            self.log.info("Pearson: %f" % (scores['pearson']))
        self.regressor_net.train()


import similarity
from similarity.normalized_levenshtein import NormalizedLevenshtein
from similarity.jarowinkler import JaroWinkler
from similarity.metric_lcs import MetricLCS
from similarity.qgram import QGram
from similarity.jaccard import Jaccard
from similarity.cosine import Cosine
from fuzzywuzzy import fuzz
from fuzzywuzzy import process
import torch

normalized_levenshtein = NormalizedLevenshtein()
jarowinkler = JaroWinkler()
metric_lcs = MetricLCS()
qgram2 = QGram(2)
qgram3 = QGram(3)
qgram4 = QGram(4)
cosine = Cosine(2)
jaccard = Jaccard(2)


def extract_string_similarity_vector(instance: dict):
    """
    Returns a vector encoding a variety of lexical similarity metrics given a dictionary containing keys
    sentence_1,sentence_2
    :return: a vector containing similarity scores
    """

    s1 = instance['sentence_1']
    s2 = instance['sentence_2']

    return torch.tensor([
        normalized_levenshtein.similarity(s1,s2),
        jarowinkler.similarity(s1,s2),
        metric_lcs.distance(s1,s2),
        qgram2.distance(s1,s2),
        qgram3.distance(s1,s2),
        qgram4.distance(s1,s2),
        jaccard.similarity(s1,s2),
        cosine.similarity(s1,s2),
        fuzz.partial_token_set_ratio(s1,s2),
        fuzz.partial_token_sort_ratio(s1,s2),
        fuzz.token_set_ratio(s1,s2),
        fuzz.token_sort_ratio(s1,s2),
        fuzz.QRatio(s1,s2),
        fuzz.UQRatio(s1,s2),
        fuzz.UWRatio(s1,s2),
        fuzz.WRatio(s1,s2)
    ])

def string_similarity_features(data: list):

    temp = extract_string_similarity_vector(data[0])

    dataset = torch.empty((len(data), temp.shape[0]), dtype=torch.float)

    for idx, instance in enumerate(data):
        dataset[idx] = extract_string_similarity_vector(instance)

    return dataset


from unittest import TestCase
from .util import get_model_path
import os

class TestSimilarityModelUtilities(TestCase):
    def test_model_download(self):
        path = get_model_path('web-bert-similarity')
        self.assertTrue(os.path.exists(path))


from .similarity import BertSimilarity
from ..util import get_model_path


class WebBertSimilarity(BertSimilarity):
    def __init__(self, device='cuda', batch_size=10, model_name="web-bert-similarity"):
        model_path = get_model_path(model_name)
        super().__init__(device=device, batch_size=batch_size, bert_model_path=model_path)


from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
from semantic_text_similarity import __version__, __authors__
import sys

packages = find_packages()

def readme():
    with open('README.md') as f:
        return f.read()

class PyTest(TestCommand):

    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)

setup(
    name='semantic_text_similarity',
    version=__version__,
    license='MIT',
    description="implementations of models and metrics for semantic text similarity. that's it.",
    long_description=readme(),
    long_description_content_type='text/markdown',
    packages=packages,
    url='https://github.com/AndriyMulyar/semantic-text-similarity',
    author=__authors__,
    author_email='contact@andriymulyar.com',
    keywords='',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Natural Language :: English',
        'Topic :: Text Processing :: Linguistic',
        'Intended Audience :: Science/Research'
    ],

    install_requires=[
        'torch',
        'strsim',
        'fuzzywuzzy[speedup]',
        'pytorch-transformers',
        'scipy'
    ],
    tests_require=["pytest"],
    cmdclass={"pytest": PyTest},
    include_package_data=True,
    zip_safe=False

)


import os, tarfile, urllib, tempfile, shutil
from pytorch_transformers.file_utils import url_to_filename, http_get

try:
    from torch.hub import _get_torch_home
    torch_cache_home = _get_torch_home()
except ImportError:
    torch_cache_home = os.path.expanduser(
        os.getenv('TORCH_HOME', os.path.join(
            os.getenv('XDG_CACHE_HOME', '~/.cache'), 'torch')))

default_cache_path = os.path.join(torch_cache_home, 'semantic_text_similarity')

MODEL_URL = {
    'clinical-bert-similarity': 'https://github.com/AndriyMulyar/semantic-text-similarity/releases/download/v1.0.0/clinical_bert_similarity.tar.gz',
    'web-bert-similarity': 'https://github.com/AndriyMulyar/semantic-text-similarity/releases/download/v1.0.0/web_bert_similarity.tar.gz'
}
def get_model_path(model_name: str):

    if model_name in MODEL_URL:
        model_url = MODEL_URL[model_name]
    else:
        return model_name
    model_url_hash = url_to_filename(model_url)

    model_path = os.path.join(default_cache_path, model_url_hash)

    if os.path.exists(model_path):
        return model_path
    else:
        os.makedirs(model_path)
        with tempfile.NamedTemporaryFile() as temp_file:
            print("Downloading model: %s from %s" % (model_name, model_url))
            try:
                http_get(model_url, temp_file)
                tar = tarfile.open(temp_file.name)
            except BaseException as exc:
                print("Failed to download model: %s"% model_name)
                os.rmdir(model_path)
                raise exc


            temp_file.flush()
            temp_file.seek(0)
            tar.extractall(model_path)

            return model_path




from pytorch_transformers import BertTokenizer
import torch


def _truncate_seq_pair(tokens_a, tokens_b, max_length: int):
    """
    Copied exactly from: https://github.com/huggingface/pytorch-pretrained-BERT/blob/78462aad6113d50063d8251e27dbaadb7f44fbf0/examples/extract_features.py#L150
    Truncates a sequence pair in place to the maximum length.
    """

    # This is a simple heuristic which will always truncate the longer sequence
    # one token at a time. This makes more sense than truncating an equal percent
    # of tokens from each, since if one sequence is very short then each token
    # that's truncated likely contains more information than a longer sequence.
    while True:
        total_length = len(tokens_a) + len(tokens_b)
        if total_length <= max_length:
            break
        if len(tokens_a) > len(tokens_b):
            tokens_a.pop()
        else:
            tokens_b.pop()


def bert_sentence_pair_preprocessing(data: list, tokenizer: BertTokenizer, max_sequence_length=128):
    """
    Pre-processes an array of sentence pairs for input into bert. Sentence pairs are expected to be processed
    as given in data.py.

    Each sentence pair is tokenized and concatenated together by the [SEP] token for input into BERT

    :return: three tensors: [data_size, input_ids], [data_size, token_type_ids], [data_size, attention_mask]
    """

    max_bert_input_length = 0
    for sentence_pair in data:

        sentence_1_tokenized, sentence_2_tokenized = tokenizer.tokenize(sentence_pair['sentence_1']), tokenizer.tokenize(sentence_pair['sentence_2'])
        _truncate_seq_pair(sentence_1_tokenized, sentence_2_tokenized, max_sequence_length - 3) #accounting for positioning tokens


        max_bert_input_length = max(max_bert_input_length, len(sentence_1_tokenized) + len(sentence_2_tokenized) + 3)
        sentence_pair['sentence_1_tokenized'] = sentence_1_tokenized
        sentence_pair['sentence_2_tokenized'] = sentence_2_tokenized

    dataset_input_ids = torch.empty((len(data), max_bert_input_length), dtype=torch.long)
    dataset_token_type_ids = torch.empty((len(data), max_bert_input_length), dtype=torch.long)
    dataset_attention_masks = torch.empty((len(data), max_bert_input_length), dtype=torch.long)
    dataset_scores = torch.empty((len(data), 1), dtype=torch.float)

    for idx, sentence_pair in enumerate(data):
        tokens = []
        input_type_ids = []

        tokens.append("[CLS]")
        input_type_ids.append(0)
        for token in sentence_pair['sentence_1_tokenized']:
            tokens.append(token)
            input_type_ids.append(0)
        tokens.append("[SEP]")
        input_type_ids.append(0)

        for token in sentence_pair['sentence_2_tokenized']:
            tokens.append(token)
            input_type_ids.append(1)
        tokens.append("[SEP]")
        input_type_ids.append(1)

        input_ids = tokenizer.convert_tokens_to_ids(tokens)

        attention_masks = [1] * len(input_ids)
        while len(input_ids) < max_bert_input_length:
            input_ids.append(0)
            attention_masks.append(0)
            input_type_ids.append(0)


        dataset_input_ids[idx] = torch.tensor(input_ids, dtype=torch.long)
        dataset_token_type_ids[idx] = torch.tensor(input_type_ids, dtype=torch.long)
        dataset_attention_masks[idx] = torch.tensor(attention_masks, dtype=torch.long)
        if 'similarity' not in sentence_pair or sentence_pair['similarity'] is None:
            dataset_scores[idx] = torch.tensor(float('nan'), dtype=torch.float)
        else:
            dataset_scores[idx] = torch.tensor(sentence_pair['similarity'], dtype=torch.float)


    return dataset_input_ids, dataset_token_type_ids, dataset_attention_masks, dataset_scores